/*  Copyright (C) 2005 Koninklijke Philips Electronics N.V.
    All Rights Reserved.
*//**
    @file   phStbDbg_Module.c
    @brief  Linux kernel module for phStbDbg component (tracing).

@b  Component:  phStbDbg.

*//*

Rev Date        Author      Comments
--------------------------------------------------------------------------------
  1 20050701    batelaan    Initial
  2 20050705    batelaan    Crosscheck fixes
  3 20050711    batelaan    Fix tmDbg_Print to check flags
  4 20050728    batelaan    Use shared memory for debug units.
  5 20050805    batelaan    Crosscheck fixes
  6 20050809    batelaan    Use separate UnitInfo struct to increase resilience against roque programs.
  7 20050824    batelaan    Crosscheck corrections. Ensure we don't overflow memory.
  8 20051122    raiyat      Linux 2.6.14.2 changes
  9 20060908    laird       Linux 2.6.17.7 changes
--------------------------------------------------------------------------------
    For consistency and standardisation retain the Section Separators.
*/

/* needed for versioned kernels (i.e. most kernels in current use) */
#include <linux/module.h>
#include <linux/kernel.h>       /* printk() */
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/spinlock.h>
#include <linux/string.h>

#include <phStbCommon.h>
#include <phStbDbg.h>
#include <phStbDbg_Internal.h>

/***********************************************
* LOCAL MACROS                                 *
* recommendation only: <MODULE><_WORD>+        *
************************************************/

#define MY_LICENSE "GPL"
MODULE_LICENSE(MY_LICENSE);

#define THIS_MODULE_DESCRIPTION "Dbg driver"
MODULE_DESCRIPTION(THIS_MODULE_DESCRIPTION);
MODULE_AUTHOR("Wouter Batelaan");

/* The number of devices we expose */
#define NR_DEVS 1

/** Maximum length of a DBG_ASSERT2 message.
    See also: tmDbgConfig.h:37:#define DBG_MAX_MSG_SIZE        512 */
#define PH_STBDBG_MAX_MSG_SIZE 512U

#if defined(UNITTEST)
  #define TRACE(...) do {printk(KERN_DEBUG __VA_ARGS__);} while (0)
#else
  #define TRACE(...) do {} while (0)
#endif

/** Print format for an unit:
    Nr: 12:<name> flags: <0x12345678>
*/
#define FORMAT "Nr: %02u:%s flags: 0x%08x\n"

/* Size of print format, without the variable name length.
   this is:
   - sizeof(FORMAT)
   - minus the 3 %specs
   - plus size of the integers printouts. */
#define FORMAT_SIZE (sizeof(FORMAT) - 10 + 2 + 8)

/******************************************************************
* FUNCTION PROTOTYPES                    [p]ph<Module>_<Word>+    *
*******************************************************************/

static int phStbDbg_open(struct inode *inode, struct file *file);
static ssize_t phStbDbg_read(struct file *file, char __user *buf, size_t count, loff_t *pos);
static ssize_t phStbDbg_release(struct inode *inode, struct file *file);
static int phStbDbg_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static ssize_t phStbDbg_mmap(struct file * file, struct vm_area_struct * vma);

/******************************************************************
* LOCAL TYPEDEFS                                                  *
* recommendation only:                   [p]ph<Module>_<Word>+_t  *
*******************************************************************/

/** Holds info (name and address of public tmDbgUnit struct) for a debug unit. */
typedef struct UnitInfo {
    struct UnitInfo *next;
    ptmDbgUnit_t pUnit;
    Char name[1];
} UnitInfo_t;
typedef UnitInfo_t *pUnitInfo_t;

/*******************************************************************
* EXPORTED DATA                 g[k|p|kp|pk|kpk]ph<Module>_<Word>+ *
********************************************************************/

#if TMDBG_PRINTS_ARE_ENABLED
DBG_UNIT(gphStbDbg_DbgUnit)
#endif

/** Public debug unit structures. These are memory-mapped into user application space. */
tmDbgUnit_t *gpphStbDbg_Entries;

/*******************************************************************
* LOCAL DATA                                                       *
********************************************************************/

static dev_t gDev; /* The start of the reserved device number region */
static struct cdev gCdev;
static struct class *gSysClass; /* The GPIO class under /sys/class */
static unsigned int gMajor; /* Major number of our device, obtained from gDev and held for convenience */
static unsigned nEntries;
static size_t gTotalNamesSize;

static pUnitInfo_t gppStbDbg_UnitInfos;
static pUnitInfo_t gppStbDbg_UnitInfoLast;

static tmDbgUnit_t *gppStbDbg_EntriesLast;
static spinlock_t gLock = SPIN_LOCK_UNLOCKED;

/* The file operations for the GPIO character devices */
static struct file_operations phStbDbg_fops =
{
    .owner      = THIS_MODULE,
    .open       = phStbDbg_open,
    .release    = phStbDbg_release,
    .read       = phStbDbg_read,
    .mmap       = phStbDbg_mmap,
    .ioctl      = phStbDbg_ioctl
};

/*******************************************************************************
* FUNCTION IMPLEMENTATION  <Module>[_<Word>+] for static functions             *
*                          tm[<layer>]<Module>[_<Word>+] for exported functions*
********************************************************************************/

/******************************************************************************/
Void phStbDbg_Lock(Void)
{
    spin_lock(&gLock);
}

/******************************************************************************/
Void phStbDbg_Unlock(Void)
{
    spin_unlock(&gLock);
}

/******************************************************************************/
UnitInfo_t *phStbDbg_New(const char *name)
{
    UnitInfo_t *pUnitInfo;
    size_t nameLen = strlen(name);

    pUnitInfo = kmalloc(sizeof(UnitInfo_t) + nameLen, GFP_KERNEL);
    if (pUnitInfo == NULL)
    {
        return NULL;
    }

    if (gppStbDbg_EntriesLast == NULL)
    {
        pUnitInfo->pUnit = gpphStbDbg_Entries;
    }
    else
    {
        /* Calculate offset from gppStbDbg_EntriesLast: */
        pUnitInfo->pUnit = (tmDbgUnit_t*)(((char*)gppStbDbg_EntriesLast) +
                                          sizeof(tmDbgUnit_t) +
                                          strlen(gppStbDbg_EntriesLast->name));
        pUnitInfo->pUnit = (tmDbgUnit_t*) ALIGN((UInt32)pUnitInfo->pUnit, sizeof(tmDbgUnit_t*));
        TRACE("phStbDbg_New (%s) pUnit=%p\n", name, pUnitInfo->pUnit);

        /* Check that there is space in the shared memory page (a singe page): */
        if (((char*)pUnitInfo->pUnit) + sizeof(tmDbgUnit_t) + nameLen - ((char*)gpphStbDbg_Entries) >= PAGE_SIZE)
        {
            kfree(pUnitInfo);
            return NULL;
        }
    }
    gppStbDbg_EntriesLast = pUnitInfo->pUnit;
    strcpy(pUnitInfo->pUnit->name, name);
    pUnitInfo->pUnit->flgs = 0;

    /* Put into linked list: */
    if (gppStbDbg_UnitInfos == NULL)
    {
        gppStbDbg_UnitInfos = pUnitInfo;
    }
    else
    {
        gppStbDbg_UnitInfoLast->next = pUnitInfo;
    }
    gppStbDbg_UnitInfoLast = pUnitInfo;
    pUnitInfo->next = NULL;

    strcpy(pUnitInfo->name, name);

    gTotalNamesSize += nameLen;
    nEntries++;
    return pUnitInfo;
}

/******************************************************************************/
UnitInfo_t *phStbDbg_Lookup(const char *name)
{
    UnitInfo_t *p = NULL;
    phStbDbg_Lock();
    for (p = gppStbDbg_UnitInfos; p != NULL; p = p->next)
    {
        TRACE("phStbDbg_Lookup (for: %s) looking at: (%s)(pUnit=%p: %s)\n",
              name, p->name, p->pUnit, p->pUnit->name);
        if (strcmp(p->name, name) == 0)
        {
            break;
        }
    }
    if (p == NULL)
    {
        p = phStbDbg_New(name);
    }
    phStbDbg_Unlock();
    return p;
}

/******************************************************************************/
#if 0
static UnitInfo_t *LookupUnitInfo(ptmDbgUnit_t pUnit)
{
    UnitInfo_t *p = NULL;
    phStbDbg_Lock();
    for (p = gppStbDbg_UnitInfos; p != NULL; p = p->next)
    {
        TRACE("LookupUnitInfo looking at: (%s)(%s)\n",
              p->name, p->pUnit->name);
        if (p->pUnit == pUnit)
        {
            TRACE("LookupUnitInfo Found it\n");
            break;
        }
    }
    phStbDbg_Unlock();
    return p;
}
#endif

/******************************************************************************/
/** Open the device.
 */
static int phStbDbg_open(struct inode *inode, struct file *file)
{
    file->private_data = NULL;

    return 0;
}

/******************************************************************************/
static ssize_t phStbDbg_release(struct inode *inode, struct file *file)
{
    if (file->private_data != NULL)
    {
        kfree(file->private_data);
    }
    return 0;
}

/******************************************************************************/
/* Return the debug flag setting. */
static char * StbDbg_GetInfo(Void)
{
    UnitInfo_t *p;
    size_t infoLen;
    size_t used = 0;
    unsigned n = 0;
    char *info = NULL;

    phStbDbg_Lock();

    infoLen = (nEntries * FORMAT_SIZE) + gTotalNamesSize;
    info = kmalloc(infoLen, GFP_KERNEL);
    if (info == NULL)
    {
        printk(KERN_WARNING "phStbDbg: kmalloc %u failed\n", infoLen);
        goto RETURN;
    }

    for (p = gppStbDbg_UnitInfos; p != NULL; p = p->next)
    {
        size_t printed;
        TRACE("phStbDbg: entry %u (%s)(%s) flags 0x%08x\n",
              n, p->name, p->pUnit->name, p->pUnit->flgs);
        printed = snprintf(
            info + used, infoLen - used,
            FORMAT,
            n, p->name, p->pUnit->flgs);
        used += printed;
        n++;
    }

  RETURN:
    phStbDbg_Unlock();
    return info;
}

/******************************************************************************/
/* Read debug flag settings. */
static ssize_t phStbDbg_read(struct file *file, char __user *buf, size_t count, loff_t *pos)
{
    ssize_t retval = 0;
    char *info;
    ssize_t len;

    if (*pos == 0)
    {
        /* Beginning of the read, so we setup the string in private_data
           to contain the debug flag setting. */
        file->private_data = StbDbg_GetInfo();
    }
    info = (char*)file->private_data;
    if (info == NULL)
    {
        return -1;
    }
    len = strlen(info);

    if (*pos < len)
    {
        if (count > len - *pos)
        {
            count = len - *pos;
        }

        if (copy_to_user(buf, info + *pos, count))
        {
            printk(KERN_WARNING "phStbDbg: copy_to_user %d bytes offset %d failed\n",
                   (int)count, (int)*pos);
            retval = -EFAULT;
        }
        else
        {
            *pos += count;
            retval = count;
        }
    }

    return retval;
}

/******************************************************************************/
/* Memory map the device. This maps ::gpphStbDbg_Entries. */
static ssize_t phStbDbg_mmap(struct file * file, struct vm_area_struct * vma)
{
    unsigned long vsize = vma->vm_end - vma->vm_start;

    if (vsize != PH_STBDBG_DEVLEN)
    {
        printk(KERN_WARNING "phStbDbg_mmap: vsize=%lu (!=%u)\n", vsize, PH_STBDBG_DEVLEN);
        return -EINVAL;
    }

    if (vma->vm_pgoff != 0)
    {
        printk(KERN_WARNING "phStbDbg_mmap: vma->vm_pgoff=%lu\n", vma->vm_pgoff);
        return -EINVAL;
    }

    if (remap_pfn_range(vma,
                        vma->vm_start,
                        ((unsigned long)gpphStbDbg_Entries) >> PAGE_SHIFT,
                        vsize,
                        vma->vm_page_prot))
        return -EAGAIN;
    return 0;
}

/******************************************************************************/
/* Handle all ioctls supported by phStbDbg */
static int phStbDbg_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    int result = -ENOTTY;

    switch (cmd)
    {
        case PH_STBDBG_IOCTL_REGISTER_MODULE:
        {
            char *name = (char*)arg;
            UnitInfo_t *pUnitInfo;
            TRACE("phStbDbg_ioctl PH_STBDBG_IOCTL_REGISTER_MODULE \"%s\"\n", name);
            pUnitInfo = phStbDbg_Lookup(name);
            if (pUnitInfo != NULL)
            {
                UInt32 offset;
                /* Convert to offset from gpphStbDbg_Entries address: */
                offset = ((char*)pUnitInfo->pUnit) - ((char*)gpphStbDbg_Entries);
                result = (int)offset;
                TRACE("phStbDbg_ioctl PH_STBDBG_IOCTL_REGISTER_MODULE \"%s\" => (%s)(%s) offset %d\n",
                      name, pUnitInfo->name, pUnitInfo->pUnit->name, result);
            }
            else
            {
                result = -ENOMEM;
            }
        }
        break;

        default:
            break;
    }

    return result;
}

/******************************************************************************/
/** Module Creation function.
    - creates the device.

    @retval int Module Initialised successfully.
*/
static int __init phStbDbg_ModuleInit(void)
{
    int result = -1;

    /** To track module initialisation progress, so we can undo correctly on error. */
    enum {
        cdevAdded = 1<<0,
        classSimpleDeviceAdded = 1<<1
    } initialised = 0;

    gpphStbDbg_Entries = (tmDbgUnit_t*)get_zeroed_page(GFP_USER);
    if (gpphStbDbg_Entries == NULL)
    {
        printk(KERN_WARNING "phStbDbg: no mem\n");
        goto ERROR_RETURN;
    }

    /* Allocate a region of minor numbers (1) under a major number */
    result = alloc_chrdev_region(&gDev, 0, NR_DEVS, PH_STBDBG_DEVNAME);
    if (result < 0)
    {
        printk(KERN_WARNING "phStbDbg: alloc_chrdev_region => %d\n", result);
        goto ERROR_RETURN;
    }

    gMajor = MAJOR(gDev);

    /* Set up /sys entry */
    gSysClass = class_create(THIS_MODULE, PH_STBDBG_DEVNAME);
    if (gSysClass == NULL)
    {
        printk(KERN_WARNING "phStbDbg: class_create => %d\n", result);
        result = -ENOMEM;
        goto ERROR_RETURN;
    }

    cdev_init(&gCdev, &phStbDbg_fops);
    gCdev.owner = THIS_MODULE;

    result = cdev_add(&gCdev, gDev, 1);
    if (result)
    {
        printk(KERN_INFO "phStbDbg: cdev_add => %d\n", result);
        goto ERROR_RETURN;
    }
    initialised |= cdevAdded;

    class_device_create(gSysClass, NULL, gDev, NULL, PH_STBDBG_DEVNAME);
    initialised |= classSimpleDeviceAdded;

    tmDbg_RegisterModule(&gphStbDbg_DbgUnit, "DBG_ROOT");
    gphStbDbg_DbgUnit->flgs = DBG_WARNING;

    printk("%s (%s-%s)\n", THIS_MODULE_DESCRIPTION, __DATE__, __TIME__);

    return 0;

  ERROR_RETURN:
    printk(KERN_WARNING "%s (%s-%s) => %d\n", THIS_MODULE_DESCRIPTION, __DATE__, __TIME__, result);
    if (gpphStbDbg_Entries != NULL)
    {
        free_page((unsigned long)gpphStbDbg_Entries);
    }

    if (initialised & classSimpleDeviceAdded)
    {
        class_device_destroy(gSysClass, gDev);
    }

    if (initialised & cdevAdded)
    {
        cdev_del(&gCdev);
    }
    return result;
}

module_init(phStbDbg_ModuleInit);

/******************************************************************************/
#if TMDBG_PRINTS_ARE_ENABLED
Void tmDbg_RegisterModule(
    ptmDbgUnit_t          * ppUnit,     // O: The registered debug unit will be
                                        //    stored here.
    const Char*             pUnitName   // I: A pointer to the name or GUID.
    )
{
    UnitInfo_t *pUnitInfo = phStbDbg_Lookup(pUnitName);
    *ppUnit = pUnitInfo->pUnit;
}
EXPORT_SYMBOL(tmDbg_RegisterModule);
#endif

/******************************************************************************/
#if TMDBG_PRINTS_ARE_ENABLED
Void tmDbg_Print(ptmDbgUnit_t uid, UInt32 flags, Char *fmt, ...)
{
    va_list args;

    TRACE(">>tmDbg_Print(\"%s\", 0x%x, \"%s\"...)\n", uid->name, flags, fmt);
    if (DBG_PRINT_CHECK(uid, flags))
    {
        Char buf[PH_STBDBG_MAX_MSG_SIZE];
        size_t len;
        char *type;
        char *priority;

        va_start(args, fmt);
        if (flags & DBG_WARNING)
        {
            priority = KERN_WARNING;
        }
        else
        {
            priority = KERN_DEBUG;
        }
        /* Add the text in front of the message, but only if one of the predefined
         * flags is enabled (and any number of custom flags). DBG_CUSTOM_FLAG - 1 is
         * exactly "all predefined flags". The switch will only find exact matches,
         * which is that only the flag is defined. */
        switch (flags & (DBG_CUSTOM_START - 1))
        {
            case DBG_STATE:
                type = "State Change ";
                break;
            case DBG_EVENTIN:
                type = "Event in: ";
                break;
            case DBG_EVENTOUT:
                type = "Event out: ";
                break;
            case DBG_INTERFACE_ENTER:
                type = "Interface enter: ";
                break;
            case DBG_INTERFACE_LEAVE:
                type = "Interface leave: ";
                break;
            case DBG_INTERNAL_ENTER:
                type = "Internal enter: ";
                break;
            case DBG_INTERNAL_LEAVE:
                type = "Internal leave: ";
                break;
            case DBG_EXTERNAL_CALL:
                type = "External call: ";
                break;
            case DBG_EXTERNAL_RETURN:
                type = "External return: ";
                break;
            case DBG_WARNING:
                type = "Warning: ";
                break;
            default:
                type = "";
        }

        len = sprintf(buf, "%s %s", uid->name, type);
        len += vsnprintf(buf + len, PH_STBDBG_MAX_MSG_SIZE - len, fmt, args);

        if (buf[len - 1] != '\n')
        {
            buf[len] = '\n';
            buf[len + 1] = '\0';
        }
        printk("%s%s", priority, buf);
    }
    TRACE("<<tmDbg_Print(\"%s\", 0x%x, \"%s\"...)\n", uid->name, flags, fmt);
}

EXPORT_SYMBOL(tmDbg_Print);
#endif


/******************************************************************************/
#if TMDBG_ASSERTS_ARE_ENABLED
Char *tmDbg_StrAssertFormat(Char *fmt, ...)
{
    static char *buf = NULL;
    va_list args;

    if (buf == NULL)
    {
        buf = kmalloc(PH_STBDBG_MAX_MSG_SIZE, GFP_KERNEL);
    }

    TRACE(">>tmDbg_StrAssertFormat(\"%s\"...)\n", fmt);
    va_start(args, fmt);
    vsnprintf(buf, PH_STBDBG_MAX_MSG_SIZE, fmt, args);
    TRACE("<<tmDbg_StrAssertFormat(\"%s\"...) => %s\n", fmt, buf);
    return buf;
}

EXPORT_SYMBOL(tmDbg_StrAssertFormat);
#endif

/******************************************************************************/
#if TMDBG_ASSERTS_ARE_ENABLED
Void tmDbg_AssertError(
    const Char *assertion,
    const Char *file,
    UInt line,
    const Char *msg)
{
    TRACE(">>tmDbg_AssertError(\"%s\", \"%s\", %u, \"%s\"\n",
          assertion, file, line, msg);
    if (msg == NULL)
    {
        msg = "(no message)";
    }
    printk(KERN_ERR "Assertion! %s: %s:%u: %s\n",
           assertion, file, line, msg);
    BUG();
    TRACE("<<tmDbg_AssertError\n");
}

EXPORT_SYMBOL(tmDbg_AssertError);
#endif //TMDBG_ASSERTS_ARE_ENABLED
/******************************************************************************/
