/*
   Copyright (C) 2005 Koninklijke Philips Electronics N.V.,
   All Rights Reserved.

Rev Date       Author      Comments
--------------------------------------------------------------------------------
  1 20050614   burningh    Original
002 20051122   raiyat      Linux 2.6.14.2 changes
003 20060908   laird       Linux 2.6.17.7 changes
004 20070216   crowe       Extended to cope with GPIO16-GPIO60 too

*/
/** GPIO (General Purpose I/O) driver for PNX8550. Handles pins 0-60. */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/proc_fs.h>

#include "phStbGpio.h"

#include <asm/mach-pnx8550/glb.h>
#include <phStbProc_BaseProcSystem.h>

/******************************************************************************
* LOCAL MACROS                                                                *
*******************************************************************************/

#define THIS_MODULE_DESCRIPTION "PNX8550 GPIO driver"

/* The base of phStbGpio device names */
#define DEVNAME "gpio"

/* The number of GPIO pins we expose as devices */
#define GPIO_NR_DEVS 61

/* Mode Control register addresses */
#define MMIO_MC(pin) (0x104000 + ((pin >> 4) << 2))

/* Obtain the current mode for the specified pin from the provided MC register */
#define PIN_MODE_READ(reg, pin) ( (reg >> ((pin % 16) * 2)) & 0x3 )

/* Set the mode for the specified pin */
#define PIN_MODE_WRITE(pin, mode) ( (mode & 0x3) << ((pin % 16) * 2) )

/* IO Data register addresses */
#define MMIO_IOD(pin) (0x104010 + ((pin >> 4) << 2))

/* Read the current data on the specified pin from the provided IOD register */
#define PIN_DATA_READ(reg, pin) ( (reg >> (pin % 16)) & 1 )

/* Write out the provided data on the specified pin */
#define PIN_DATA_WRITE(pin, val) ( (1 << (16 + (pin % 16))) + (val << (pin % 16)) )

/* Set the given pin to tristate */
#define PIN_TRISTATE(pin) (1 << (pin % 16))

/* Determine the length of an array */
#define ARRAYLEN(x) (sizeof(x) / sizeof(*x))

/* The length of internal buffers used for read and write operations */
#define KERN_BUFFER_LEN 200

/* Module set up */
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION(THIS_MODULE_DESCRIPTION);
MODULE_AUTHOR("Neil Burningham");

/******************************************************************************
* STATIC FUNCTION PROTOTYPES                                                  *
*******************************************************************************/

static int phStbGpio_open(struct inode *inode, struct file *file);
static ssize_t phStbGpio_read(struct file *file, char __user *buf, size_t count, loff_t *pos);
static ssize_t phStbGpio_write(struct file *file, const char __user *buf, size_t count, loff_t *pos);
static int phStbGpio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);

/******************************************************************************
* LOCAL TYPEDEFS                                                              *
*******************************************************************************/

/* Data held by a driver instance */
struct phStbGpio_t
{
    dev_t devRegion; /* The start of the reserved GPIO device number region */
    struct cdev cdev[GPIO_NR_DEVS]; /* The array of GPIO character devices */
    struct class *sysClass; /* The GPIO class under /sys/class */
    unsigned int major; /* Major number of our device, obtained from devRegion and held for convenience */
    struct proc_dir_entry *procDir; /* Our directory under /proc */
};

/* Structure for number to string lookup tables */
struct lookup_t
{
    int number;
    char *string;
};

/******************************************************************************
* EXPORTED DATA                                                               *
*******************************************************************************/

/******************************************************************************
* LOCAL DATA                                                                  *
*******************************************************************************/

/* The driver instance data */
static struct phStbGpio_t gInstance;


/* Lookup table for modes of a GPIO pin */
static struct lookup_t mode_lookup[] =
{
    { phStbGpio_PinModePrimary,       "Primary pin mode (non GPIO)" },
    { phStbGpio_PinModeGpio,          "GPIO mode (standard)"        },
    { phStbGpio_PinModeGpioOpenDrain, "GPIO mode (open drain)"      }
};


/* The file operations for the GPIO character devices */
static struct file_operations phStbGpio_fops =
{
    .owner      = THIS_MODULE,
    .open       = phStbGpio_open,
    .release    = NULL,
    .read       = phStbGpio_read,
    .write      = phStbGpio_write,
    .ioctl      = phStbGpio_ioctl
};

/******************************************************************************
* FUNCTION IMPLEMENTATION                                                     *
*******************************************************************************/

/* Look up a mode and convert it into a textual description */
static char *phStbGpio_lookupMode(int mode)
{
    char *retval = NULL;
    int i;

    for (i = 0; i < ARRAYLEN(mode_lookup); i++)
    {
        struct lookup_t *lookup = &mode_lookup[i];

        if (lookup->number == mode)
        {
            retval = lookup->string;
        }
    }

    return retval;
}


/* Open the device - put the pin number into the file structure */
static int phStbGpio_open(struct inode *inode, struct file *file)
{
    /* Minor number == pin number */
    file->private_data = (void *)iminor(inode);

    return 0;
}


/* Output current value on GPIO pin, 0 or 1 */
static ssize_t phStbGpio_read(struct file *file, char __user *buf, size_t count, loff_t *pos)
{
    ssize_t retval = 0;

    unsigned int pin = (int)file->private_data;
    unsigned long iod;
    unsigned long data;
    char kernBuf[KERN_BUFFER_LEN];
    ssize_t len;
    
    iod = MMIO(MMIO_IOD(pin));
    
    data = PIN_DATA_READ(iod, pin);
    
    len = sprintf(kernBuf, "%ld\n", data);
    
    if (count > len)
    {
	count = len;
    }
    
    if ( copy_to_user(buf, kernBuf, count) )
    {
	printk(KERN_WARNING "%s: Unable to copy_to_user all data from sprintf during read.\n", __FILE__);
	retval = -EFAULT;
    }
    else
    {
	*pos += count;
	retval = count;
    }

    return retval;
}


/* Write 0 or 1 to set pin value */
static ssize_t phStbGpio_write(struct file *file, const char __user *buf, size_t count, loff_t *pos)
{
    ssize_t retval = 0;

    char kernBuf[KERN_BUFFER_LEN];
    
    if (count > KERN_BUFFER_LEN)
    {
	count = KERN_BUFFER_LEN;
    }
    
    if ( copy_from_user(kernBuf, buf, count) )
    {
	printk(KERN_WARNING "%s: Unable to copy_from_user all data during write.\n", __FILE__);
	retval = -EFAULT;
    }
    else
    {
	size_t index = 0;
	while (index < count)
	{
	    if ((kernBuf[index] >= '0') && (kernBuf[index] <= '1'))
	    {
		unsigned int pin = (int)file->private_data;
		unsigned long iod;
		
		kernBuf[index] -= '0'; /* Translate from ASCII to decimal */
		
		iod = PIN_DATA_WRITE(pin, kernBuf[index]);
		
		MMIO(MMIO_IOD(pin)) = iod;
	    }
	    else
	    {
		printk(KERN_INFO "%s: Received unknown value %c.\n", __FILE__, kernBuf[0]);
	    }
	    ++index;
	}
	
	retval = count;
	*pos += count;
    }

    return retval;
}


/* Handle the PHSTBGPIO_IOCTL_SET_PIN_MODE ioctl */
static int phStbGpio_SetPinMode(unsigned int pin, unsigned long arg)
{
    phStbGpio_PinMode_t mode = (phStbGpio_PinMode_t)arg;
    unsigned long mc;

    /* Handle modifications to MC register */
    switch (mode)
    {
        case phStbGpio_PinModePrimary:
        case phStbGpio_PinModeGpioOpenDrain:
            mc = PIN_MODE_WRITE(pin, mode);
            break;

        case phStbGpio_PinModeGpio:
        case phStbGpio_PinModeGpioTristate:
            mc = PIN_MODE_WRITE(pin, phStbGpio_PinModeGpio);
            break;

        default:
            return -EINVAL;
            break;
    }

    MMIO(MMIO_MC(pin)) = mc;

    /* Handle modifications to IOD register if necessary */
    if (mode == phStbGpio_PinModeGpioTristate)
    {
        unsigned long iod;
        iod = PIN_TRISTATE(pin);

        MMIO(MMIO_IOD(pin)) = iod;
    }

    return 0;
}


/* Handle the PHSTBGPIO_IOCTL_GET_PIN_MODE ioctl */
static int phStbGpio_GetPinMode(unsigned int pin, unsigned long arg)
{
    phStbGpio_PinMode_t *pMode = (phStbGpio_PinMode_t *)arg;
    unsigned long mc;

    mc = MMIO(MMIO_MC(pin));

    *pMode = PIN_MODE_READ(mc, pin);

    return 0;
}


/* Handle all ioctls supported by phStbGpio */
static int phStbGpio_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    unsigned int pin = (int)file->private_data;
    int result = -ENOTTY;

    switch (cmd)
    {
        case PHSTBGPIO_IOCTL_SET_PIN_MODE:
            result = phStbGpio_SetPinMode(pin, arg);
            break;

        case PHSTBGPIO_IOCTL_GET_PIN_MODE:
            result = phStbGpio_GetPinMode(pin, arg);

        default:
            break;
    }

    return result;
}


/* Handle proc reads */
int phStbGpio_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    int len = 0;

    if (offset == 0)
    {
        unsigned int pin = (int)data;
        unsigned long mc;
        unsigned long iod;
        unsigned long mode, data;

        mc = MMIO(MMIO_MC(pin));
        iod = MMIO(MMIO_IOD(pin));

        mode = PIN_MODE_READ(mc, pin);
        data = PIN_DATA_READ(iod, pin);

        len = sprintf(buf, "Pin %d:\n\tMode %ld [ %s ] \n\tData %ld\n", pin, mode, phStbGpio_lookupMode(mode), data);

        *eof = 1;
    }

    return len;
}


/* Create a character device for the specified index, add a /sys and /proc entry */
static int phStbGpio_CreateChrDev(unsigned int index)
{
    int result;
    dev_t devno = MKDEV(gInstance.major, index); /* Assumes that base is 0 */
    char devname[20];

    cdev_init(&gInstance.cdev[index], &phStbGpio_fops);
    gInstance.cdev[index].owner = THIS_MODULE;

    result = cdev_add (&gInstance.cdev[index], devno, 1);
    if (result)
    {
        printk(KERN_INFO "%s: Failed to add character device (0x%X)\n", __FILE__, result);
        return result;
    }

    sprintf(devname, "%s%d", DEVNAME, index);
    class_device_create(gInstance.sysClass, NULL, devno, NULL, devname);

    sprintf(devname, "%d", index);
    create_proc_read_entry(devname, 0, gInstance.procDir, phStbGpio_read_proc, (void *)index);

    return 0;
}


/* Delete a character device for the specified index, remove corresponding /sys and /proc entries */
static void phStbGpio_DelChrDev(unsigned int index)
{
    dev_t devno = MKDEV(gInstance.major, index);
    char devname[20];

    cdev_del(&gInstance.cdev[index]);

    class_device_destroy(gInstance.sysClass, devno);

    sprintf(devname, "%d", index);
    remove_proc_entry(devname, gInstance.procDir);
}


/* Initialise the module - create all devices */
static int __init phStbGpio_init(void)
{
    struct proc_dir_entry *baseDir;
    int result;
    int i;

    printk ("%s (%s-%s)\n", THIS_MODULE_DESCRIPTION, __DATE__, __TIME__);

    /* Allocate a region of minor numbers under a major number for our GPIO device */
    result = alloc_chrdev_region(&gInstance.devRegion, 0, GPIO_NR_DEVS, DEVNAME);
    if (result < 0)
    {
        printk(KERN_INFO "%s: alloc_chrdev_region failed %d\n", __FILE__, result);
        return result;
    }

    gInstance.major = MAJOR(gInstance.devRegion);

    /* Set up /sys and /proc entries */
    gInstance.sysClass = class_create(THIS_MODULE, DEVNAME);
    if (gInstance.sysClass == NULL)
    {
        printk(KERN_INFO "%s: class_create failed %d\n", __FILE__, result);
        return -ENOMEM;
    }

    result = phStbProc_GetPNX8550ProcEntry(&baseDir);
    if (result < 0)
    {
        printk(KERN_INFO "%s: phStbProc_GetPNX8550ProcEntry failed %d\n", __FILE__, result);
        return result;
    }

    gInstance.procDir = proc_mkdir("gpio", baseDir);
    if (gInstance.procDir == NULL)
    {
        printk(KERN_INFO "%s: proc_mkdir failed\n", __FILE__);
        return -ENOMEM;
    }

    /* Create the actual character devices */
    for (i = 0; (i < GPIO_NR_DEVS) && (result == 0); i++)
    {
        result = phStbGpio_CreateChrDev(i);
    }

    return result;
}


/* Deinitialise the module - destroy all devices */
static void __exit phStbGpio_exit(void)
{
    struct proc_dir_entry *baseDir;
    int i;

    printk(KERN_INFO "Shutting down %s\n", THIS_MODULE_DESCRIPTION);

    for (i = 0; i < GPIO_NR_DEVS; i++)
    {
        phStbGpio_DelChrDev(i);
    }

    class_destroy(gInstance.sysClass);

    phStbProc_GetPNX8550ProcEntry(&baseDir);
    remove_proc_entry("gpio", baseDir);

    unregister_chrdev_region(gInstance.devRegion, GPIO_NR_DEVS);
}


module_init(phStbGpio_init);
module_exit(phStbGpio_exit);

