/*
                                  IPSTB Project
                         ---------------------------------


   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Copyright (C) 2005 Koninklijke Philips Electronics N.V.,
   All Rights Reserved.

Filename: phStbRTC.c


Rev Date       Author      Comments
--------------------------------------------------------------------------------
  1 20050525   Mike Neill  Initial.
  2 20050531   batelaan    Updated copyright.
  3 20050602   Mike Neill  Corrected date display
  4 20050705   Mike Neill  Improved integration with 'struct tm' format
  5 20050705   Mike Neill  Crosscheck corrections
  8 20050802   batelaan    Ensure year is not > 2037, this causes Linux time wraparound problems.
                           On module init check and reset if needed to 01/01/2005.
  9 20051122   raiyat      Linux 2.6.14.2 changes
 10 20060809   laird       Linux 2.6.17.7 changes

*/

/** @file
    RTC driver for PCF8563.
*/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/kmod.h>
#include <linux/bcd.h>
#include <linux/time.h>
#include <asm/rtc.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#define THIS_MODULE_DESCRIPTION "RTC driver"
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION(THIS_MODULE_DESCRIPTION);
MODULE_AUTHOR("Mike Neill");

#include <linux/i2c.h>
#include <linux/i2c-dev.h>

#include <linux/miscdevice.h>

#define RTC_CONTROL1                0x00
#define RTC_CONTROL2                0x01
#define RTC_SECONDS                 0x02
#define RTC_MINUTES                 0x03
#define RTC_HOURS                   0x04
#define RTC_DAY_OF_MONTH            0x05
#define RTC_WEEKDAY                 0x06
#define RTC_MONTH_CENTURY           0x07
#define RTC_YEAR                    0x08
#define RTC_MINUTE_ALARM            0x09
#define RTC_HOUR_ALARM              0x0A
#define RTC_DAY_ALARM               0x0B
#define RTC_WEEKDAY_ALARM           0x0C
#define RTC_CLOCKOUT_FREQ           0x0D
#define RTC_TIMER_CONTROL           0x0E
#define RTC_TIMER_COUNTDOWN         0x0F

#define CONTROL1_MASK               0xA8
#define CONTROL2_MASK               0x1F
#define CONTROL2_AF_MASK            0x08
#define SECONDS_MASK                0x7F
#define LOW_VOLTAGE_MASK            0x80
#define MINUTES_MASK                0x7F
#define HOURS_MASK                  0x3F
#define DAYS_MASK                   0x3F
#define WEEKDAYS_MASK               0x07
#define MONTHS_MASK                 0x1F
#define CENTURY_MASK                0x80
#define YEARS_MASK                  0xFF
#define ALARM_ENABLE_MASK           0x80
#define CLOCKOUT_FREQ_MASK          0x83
#define TIMER_CONTROL_MASK          0x83
#define COUNTDOWN_MASK              0xFF

#define ALARM_INTERRUPT_ENABLE_MASK 0x02
#define ALARM_PENDING_MASK          0x08

#define RTC_I2C_BUS                 3

#define RTC_MINOR_NUMBER            135

#define I2C_NAME(s)                 (s)->name

#define I2C_PCF8563                 0xA2

static int debug = 0;
module_param(debug, int, 0);
MODULE_PARM_DESC(debug, "Debug level (0-1)");

#define dprintk(num, format, args...) \
        do { \
                if (debug >= num) \
                        printk(format, ##args); \
        } while (0)

static int phStbRTC_command (     struct inode  *inode, struct file   *file,
                                unsigned int   cmd, unsigned long  arg);
static struct file_operations phStbRTC_fops = {
        .owner          = THIS_MODULE,
        .read           = NULL,
        .open           = NULL,
        .release        = NULL,
        .ioctl          = phStbRTC_command,
};

static int phStbRTC_initialised;
static struct i2c_driver i2c_driver_phStbRTC;
static struct i2c_client *phStbRTC_client;
static struct miscdevice phStbRTCMisc = { RTC_MINOR_NUMBER, "rtc", &phStbRTC_fops };

/*
 * Generic i2c probe
 * concerning the addresses: i2c wants 7 bit (without the r/w bit), so '>>1'
 */
static unsigned short normal_i2c[] = {
        I2C_PCF8563 >> 1,
        I2C_CLIENT_END
};

static int phStbRTC_attach_adapter (struct i2c_adapter *adapter);
static int phStbRTC_detach_client (struct i2c_client *client);

static unsigned short ignore = I2C_CLIENT_END;

static struct i2c_client_address_data addr_data = {
	.normal_i2c		= normal_i2c,
	.probe			= &ignore,
	.ignore			= &ignore,
};

static struct i2c_driver i2c_driver_phStbRTC = {
        .driver = {
		    .name = "phStbRTC",
	    },
        .id = I2C_DRIVERID_PCF8563,
        .attach_adapter = phStbRTC_attach_adapter,
        .detach_client = phStbRTC_detach_client,
};

static const u8 days_in_month[] =
    { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

/* -------------------------- I2C Support Functions --------------------------- */

static inline int phStbRTC_write_reg ( const u8        reg,
                                         const u8        data)
{
    dprintk(1,
            KERN_INFO
            "%s: write register %d  Value %02x\n", THIS_MODULE_DESCRIPTION, reg, data);
    return i2c_smbus_write_byte_data(phStbRTC_client, reg, data);
}

static inline int phStbRTC_read_reg ( const u8 reg)
{
    int ret_value;

    ret_value = i2c_smbus_read_byte_data(phStbRTC_client, reg);

    if (ret_value >= 0)
    {
        /* The PCF8563 does not return 0 for unimplemented bits */
        switch(reg)
        {
            case RTC_CONTROL1 :
                 ret_value &= CONTROL1_MASK;
                break;
            case RTC_CONTROL2 :
                 ret_value &= CONTROL2_MASK;
                break;
            case RTC_SECONDS :
                 ret_value &= LOW_VOLTAGE_MASK|SECONDS_MASK;
                break;
            case RTC_MINUTES :
                 ret_value &= MINUTES_MASK;
                break;
            case RTC_HOURS :
                 ret_value &= HOURS_MASK;
                break;
            case RTC_DAY_OF_MONTH :
                 ret_value &= DAYS_MASK;
                break;
            case RTC_WEEKDAY :
                 ret_value &= WEEKDAYS_MASK;
                break;
            case RTC_MONTH_CENTURY :
                 ret_value &= CENTURY_MASK|MONTHS_MASK;
                break;
            case RTC_YEAR :
                 ret_value &= YEARS_MASK;
                break;
            case RTC_MINUTE_ALARM :
                 ret_value &= ALARM_ENABLE_MASK|MINUTES_MASK;
                break;
            case RTC_HOUR_ALARM :
                 ret_value &= ALARM_ENABLE_MASK|HOURS_MASK;
                break;
            case RTC_DAY_ALARM :
                 ret_value &= ALARM_ENABLE_MASK|DAYS_MASK;
                break;
            case RTC_WEEKDAY_ALARM :
                 ret_value &= ALARM_ENABLE_MASK|WEEKDAYS_MASK;
                break;
            case RTC_CLOCKOUT_FREQ :
                 ret_value &= CLOCKOUT_FREQ_MASK;
                break;
            case RTC_TIMER_CONTROL :
                 ret_value &= TIMER_CONTROL_MASK;
                break;
            case RTC_TIMER_COUNTDOWN :
                 ret_value &= COUNTDOWN_MASK;
                break;
        }
    }
    else
    {
        dprintk(1,
                KERN_ERR
                "%s: Error reading RTC register %d (%d)\n", THIS_MODULE_DESCRIPTION, reg, ret_value);
    }
    dprintk(1,
            KERN_INFO
            "%s: read register %d  Value %02x\n", THIS_MODULE_DESCRIPTION, reg, ret_value);
    return ret_value;
}

static void phStbRTC_get_rtc_time(struct rtc_time *tm)
{
    if (phStbRTC_initialised == 0)
    {
        memset(tm, 0, sizeof (struct rtc_time));
        dprintk(1,
                KERN_ERR
                "%s: get_rtc_time - module not initialised\n", THIS_MODULE_DESCRIPTION);
        return;
    }
    tm->tm_sec = phStbRTC_read_reg(RTC_SECONDS);
    tm->tm_min = phStbRTC_read_reg(RTC_MINUTES) & MINUTES_MASK;
    tm->tm_hour = phStbRTC_read_reg(RTC_HOURS) & HOURS_MASK;
    tm->tm_mday = phStbRTC_read_reg(RTC_DAY_OF_MONTH) & DAYS_MASK;
    tm->tm_mon = phStbRTC_read_reg(RTC_MONTH_CENTURY);
    tm->tm_year = phStbRTC_read_reg(RTC_YEAR) & YEARS_MASK;

    if (tm->tm_sec & LOW_VOLTAGE_MASK)
    {
        printk(KERN_WARNING "%s: Low Voltage - date/time is not reliable!\n", THIS_MODULE_DESCRIPTION);
        /* Reset the VL flag */
        phStbRTC_write_reg(RTC_SECONDS, tm->tm_sec&SECONDS_MASK);
    }

    BCD_TO_BIN(tm->tm_year);
    if (tm->tm_mon & CENTURY_MASK)
    {
        tm->tm_year += 100;
    }

    tm->tm_sec &= SECONDS_MASK;
    tm->tm_mon &= MONTHS_MASK;

    BCD_TO_BIN(tm->tm_sec);
    BCD_TO_BIN(tm->tm_min);
    BCD_TO_BIN(tm->tm_hour);
    BCD_TO_BIN(tm->tm_mday);
    BCD_TO_BIN(tm->tm_mon);
    tm->tm_mon--;
}

static void phStbRTC_set_system_time(struct rtc_time *tm)
{
    struct timespec tv;
    tv.tv_sec = mktime(tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
    tv.tv_nsec = 0;
    do_settimeofday(&tv);
}

static int phStbRTC_set_rtc_time(struct rtc_time *tm)
{
    struct rtc_time local_time;
    int leap;
    int century;

    if (phStbRTC_initialised == 0)
    {
        dprintk(1,
                KERN_ERR
                "%s: set_rtc_time - module not initialised\n", THIS_MODULE_DESCRIPTION);
        return -EINVAL;
    }

    local_time = *tm;

    leap = ((local_time.tm_mon == 1) && ((local_time.tm_year % 4) == 0)) ? 1 : 0;

    /* Perform some sanity checks. */
    if ((local_time.tm_year < 70) ||
        (local_time.tm_year > 137) ||
        (local_time.tm_mon > 11) ||
        (local_time.tm_mday == 0) ||
        (local_time.tm_mday > days_in_month[local_time.tm_mon] + leap) ||
        (local_time.tm_hour >= 24) ||
        (local_time.tm_min >= 60) ||
        (local_time.tm_sec >= 60))
        return -EINVAL;

    phStbRTC_set_system_time(tm);

    century = (local_time.tm_year >= 100) ? CENTURY_MASK : 0;
    local_time.tm_year = local_time.tm_year % 100;

    /* Allow for the storage of months as 1-12 not 0-11! */
    local_time.tm_mon++;

    BIN_TO_BCD(local_time.tm_year);
    BIN_TO_BCD(local_time.tm_mon);
    BIN_TO_BCD(local_time.tm_mday);
    BIN_TO_BCD(local_time.tm_hour);
    BIN_TO_BCD(local_time.tm_min);
    BIN_TO_BCD(local_time.tm_sec);
    local_time.tm_mon |= century;

    phStbRTC_write_reg(RTC_YEAR, local_time.tm_year);
    phStbRTC_write_reg(RTC_MONTH_CENTURY, local_time.tm_mon);
    phStbRTC_write_reg(RTC_DAY_OF_MONTH, local_time.tm_mday);
    phStbRTC_write_reg(RTC_HOURS, local_time.tm_hour);
    phStbRTC_write_reg(RTC_MINUTES, local_time.tm_min);
    phStbRTC_write_reg(RTC_SECONDS, local_time.tm_sec);

    return 0;
}

static void phStbRTC_set_rtc_alarm_on(void)
{
    u8 control;

    /* Clear any pending alarms */
    control = phStbRTC_read_reg(RTC_CONTROL2) & ~ALARM_PENDING_MASK;
    phStbRTC_write_reg(RTC_CONTROL2, control|ALARM_INTERRUPT_ENABLE_MASK);
}

static void phStbRTC_set_rtc_alarm_off(void)
{
    u8 control;

    /* Clear any pending alarms */
    control = phStbRTC_read_reg(RTC_CONTROL2) & ~(ALARM_PENDING_MASK | ALARM_INTERRUPT_ENABLE_MASK);
    phStbRTC_write_reg(RTC_CONTROL2, control);
}

static void phStbRTC_get_rtc_alarm(struct rtc_time *time)
{
    u8 minute;
    u8 hour;
    u8 day;

    minute = phStbRTC_read_reg(RTC_MINUTE_ALARM) & MINUTES_MASK;
    hour = phStbRTC_read_reg(RTC_HOUR_ALARM) & HOURS_MASK;
    day = phStbRTC_read_reg(RTC_DAY_ALARM) & DAYS_MASK;

    time->tm_min = BCD_TO_BIN(minute);
    time->tm_hour = BCD_TO_BIN(hour);
    time->tm_mday = BCD_TO_BIN(day);
}

static void phStbRTC_get_rtc_alarm_pending(unsigned char *pending)
{
    *pending = (phStbRTC_read_reg(RTC_CONTROL2) & ALARM_PENDING_MASK) ? 1 : 0;
}

static void phStbRTC_get_rtc_alarm_enabled(unsigned char *enabled)
{
    *enabled = (phStbRTC_read_reg(RTC_CONTROL2) & ALARM_INTERRUPT_ENABLE_MASK) ? 1 : 0;
}

static int phStbRTC_set_rtc_alarm(struct rtc_time *time)
{
    u8 next_month;
    u8 enabled;
    int leap;
    int next_leap;
    struct rtc_time current_time;
    struct rtc_time local_time;

    local_time = *time;

    phStbRTC_get_rtc_time(&current_time);
    next_month = (current_time.tm_mon+1)%12;

    leap = ((current_time.tm_mon == 1) && ((current_time.tm_year % 4) == 0)) ? 1 : 0;
    next_leap = ((next_month == 1) && ((current_time.tm_year % 4) == 0)) ? 1 : 0;

    /* Perform some sanity checks. */
    if ((local_time.tm_mday == 0) ||
        (local_time.tm_hour >= 24) ||
        (local_time.tm_min >= 60) ||
        ((local_time.tm_mday >= current_time.tm_mday) && (local_time.tm_mday > days_in_month[current_time.tm_mon]+leap)) ||
        ((local_time.tm_mday < current_time.tm_mday)  && (local_time.tm_mday > days_in_month[next_month]+next_leap)))
        return -EINVAL;

    BIN_TO_BCD(local_time.tm_mday);
    BIN_TO_BCD(local_time.tm_hour);
    BIN_TO_BCD(local_time.tm_min);

    phStbRTC_get_rtc_alarm_enabled(&enabled);

    /* Disable the interrupt */
    phStbRTC_set_rtc_alarm_off();

    /* Disable the weekday alarm as it is not used - set the Alarm Enable bit! :-) */
    phStbRTC_write_reg(RTC_WEEKDAY_ALARM, ALARM_ENABLE_MASK);

    /* Set the minute, hour and day values - clear the Alarm Enable bit to enable! :-) */
    phStbRTC_write_reg(RTC_MINUTE_ALARM, local_time.tm_min);
    phStbRTC_write_reg(RTC_HOUR_ALARM, local_time.tm_hour);
    phStbRTC_write_reg(RTC_DAY_ALARM, local_time.tm_mday);

    if (enabled)
    {
        phStbRTC_set_rtc_alarm_on();
    }

    return 0;
}

static int phStbRTC_command ( struct inode  *inode, struct file   *file,
                              unsigned int   cmd,   unsigned long  arg)
{
    dprintk(1, KERN_INFO "%s: Got command : 0x%X\n", THIS_MODULE_DESCRIPTION, cmd);
    switch (cmd)
    {
        case RTC_RD_TIME:
        {
            struct rtc_time tm;

            memset(&tm, 0, sizeof (struct rtc_time));

            phStbRTC_get_rtc_time(&tm);

            if (copy_to_user((struct rtc_time *) arg, &tm, sizeof(struct rtc_time))) {
                return -EFAULT;
            }
        }
        break;
        case RTC_SET_TIME:
        {
            struct rtc_time tm;

            memset(&tm, 0, sizeof (struct rtc_time));

            if (copy_from_user(&tm, (struct rtc_time *) arg, sizeof(struct rtc_time)))
                return -EFAULT;

            return phStbRTC_set_rtc_time(&tm);
        }
        break;
        case RTC_ALM_READ:
        {
            struct rtc_time tm;

            memset(&tm, 0, sizeof (struct rtc_time));

            phStbRTC_get_rtc_alarm(&tm);

            if (copy_to_user((struct rtc_time *) arg, &tm, sizeof(struct rtc_time))) {
                return -EFAULT;
            }
        }
        break;
        case RTC_ALM_SET:
        {
            struct rtc_time tm;

            memset(&tm, 0, sizeof (struct rtc_time));

            if (copy_from_user(&tm, (struct rtc_time *) arg, sizeof(struct rtc_time)))
                return -EFAULT;

            return phStbRTC_set_rtc_alarm(&tm);
        }
        break;
        case RTC_AIE_ON:
        {
            phStbRTC_set_rtc_alarm_on();
        }
        break;
        case RTC_AIE_OFF:
        {
            phStbRTC_set_rtc_alarm_off();
        }
        break;
        case RTC_WKALM_SET:
        {
            struct rtc_wkalrm wkalrm;
            int ret_value = 0;

            memset(&wkalrm, 0, sizeof (struct rtc_wkalrm));

            if (copy_from_user(&wkalrm, (struct rtc_wkalrm *) arg, sizeof(struct rtc_wkalrm)))
                return -EFAULT;

            if (wkalrm.enabled)
            {
                ret_value = phStbRTC_set_rtc_alarm(&wkalrm.time);
                if (ret_value == 0)
                {
                    phStbRTC_set_rtc_alarm_on();
                }
            }
            else
            {
                phStbRTC_set_rtc_alarm_off();
            }
            return ret_value;
        }
        break;
        case RTC_WKALM_RD:
        {
            struct rtc_wkalrm wkalrm;

            memset(&wkalrm, 0, sizeof (struct rtc_wkalrm));

            phStbRTC_get_rtc_alarm(&wkalrm.time);
            phStbRTC_get_rtc_alarm_pending(&wkalrm.pending);
            phStbRTC_get_rtc_alarm_enabled(&wkalrm.enabled);

            if (copy_to_user((struct rtc_wkalrm *) arg, &wkalrm, sizeof(struct rtc_wkalrm))) {
                return -EFAULT;
            }
        }
        break;

        default:
        {
            dprintk(1, KERN_ERR "%s: unknown cmd(%d)\n",
                    THIS_MODULE_DESCRIPTION, cmd);
            return -EINVAL;
        }
    }
    return 0;
}

/* -------------------------- I2C Detection Functions ------------------------- */

static int phStbRTC_detect_client (struct i2c_adapter *adapter, int address, int kind)
{
    int rv;

    dprintk(1,
            KERN_INFO
            "%s: detecting phStbRTC client on address 0x%x (adapter %d)\n",
            THIS_MODULE_DESCRIPTION, address << 1, adapter->nr);

    /* Check to see if the RTC device has already been detected */
    if (phStbRTC_initialised)
    {
        return 0;
    }

    /* Check that the RTC device has been found on the correct I2C bus - address clash with EEPROM! */
    if (adapter->nr != RTC_I2C_BUS)
    {
        return 0;
    }

    /* Check if the adapter supports the needed features */
    if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_READ_BYTE | I2C_FUNC_SMBUS_WRITE_BYTE_DATA))
    {
        return 0;
    }

    /* Create the I2C client data structure */
    phStbRTC_client = kzalloc(sizeof(struct i2c_client), GFP_KERNEL);
    if (phStbRTC_client == 0)
    {
        return -ENOMEM;
    }
    phStbRTC_client->addr = address;
    phStbRTC_client->adapter = adapter;
    phStbRTC_client->driver = &i2c_driver_phStbRTC;
    snprintf(I2C_NAME(phStbRTC_client), sizeof(I2C_NAME(phStbRTC_client)) - 1, "phStbRTC");

    /* Attach the RTC to the I2C adapter */
    rv = i2c_attach_client(phStbRTC_client);
    if (rv) {
        dprintk(1,
                KERN_ERR
                "%s: Failed to attach to client on adapter %s (0x%x) - 0x%02X\n",
                THIS_MODULE_DESCRIPTION,
                I2C_NAME(adapter), adapter->id, rv);
        kfree(phStbRTC_client);
        return rv;
    }

    if (misc_register(&phStbRTCMisc)!=0)
    {
        printk(KERN_WARNING "%s: could not misc_register the device\n", THIS_MODULE_DESCRIPTION);
    }

    phStbRTC_initialised = 1;

    /* Get the current time from the RTC */
    {
        struct rtc_time tm;

        phStbRTC_get_rtc_time(&tm);

        printk( KERN_INFO "%s: System time is %02d/%02d/%04d %02d:%02d:%02d\n",
                THIS_MODULE_DESCRIPTION,
                tm.tm_mday, tm.tm_mon+1, tm.tm_year+1900, tm.tm_hour, tm.tm_min, tm.tm_sec);

        if (tm.tm_year < 105 || tm.tm_year > 137)
        {
            printk( KERN_WARNING "%s: System time out of range; setting to 01/01/2005\n", THIS_MODULE_DESCRIPTION);
            tm.tm_year = 105;
            tm.tm_mday = 1;
            tm.tm_mon = tm.tm_hour = tm.tm_min = tm.tm_sec = 0;
            phStbRTC_set_rtc_time(&tm); // also calls phStbRTC_set_system_time
        }
        else
        {
            phStbRTC_set_system_time(&tm);
        }
    }

    dprintk(1,
            KERN_INFO
            "%s: found device on adapter %s (0x%x) - 0x%02X\n",
            THIS_MODULE_DESCRIPTION,
            I2C_NAME(adapter), adapter->id, rv);

    return 0;
}

static int phStbRTC_attach_adapter (struct i2c_adapter *adapter)
{
        return i2c_probe(adapter, &addr_data, &phStbRTC_detect_client);
}

static int phStbRTC_detach_client (struct i2c_client *client)
{
        int err;

        err = i2c_detach_client(client);
        if (err)
        {
                return err;
        }

        kfree(client);

        return 0;
}

/* -------------------------- MODULE INIT / EXIT ------------------------------ */

static int __init phStbRTC_init (void)
{
        int ret;
        phStbRTC_initialised = 0;

        ret = i2c_add_driver(&i2c_driver_phStbRTC);
        if (ret == 0)
        {
            printk ("%s (%s-%s)\n", THIS_MODULE_DESCRIPTION, __DATE__, __TIME__);
        }
        return ret;
}

static void __exit phStbRTC_exit (void)
{
        phStbRTC_initialised = 0;
    if (misc_deregister(&phStbRTCMisc)!=0)
    {
                printk(KERN_WARNING "%s: could not misc_deregister the device\n", THIS_MODULE_DESCRIPTION);
    }
        i2c_del_driver(&i2c_driver_phStbRTC);
        printk("Shutting down %s\n", THIS_MODULE_DESCRIPTION);
}

module_init(phStbRTC_init);
module_exit(phStbRTC_exit);
