/*  Copyright (C) 2004 Koninklijke Philips Electronics N.V.
    All Rights Reserved.
*//**
    @file   phStbRpc_LudditeXferModule.c
    @brief  Xfer RPC functions using Luddite for Linux kernel phStbRpc module.

@b  Component:  phStbRpc.

*//*

Rev Date        Author      Comments
--------------------------------------------------------------------------------
  1 20051109    batelaan    Initial
--------------------------------------------------------------------------------
    For consistency and standardisation retain the Section Separators.
*/

/* needed for versioned kernels (i.e. most kernels in current use) */
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
//#include <linux/proc_fs.h>
#include <linux/slab.h>        /* kmalloc() */ 
#include <linux/errno.h>
#include <linux/fs.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <linux/delay.h>

#include <linux/device.h>
#include <linux/cdev.h>

#include <phStbRpc.h>
#include <phStbRpc_Internal.h>
#include <phStbRpc_Xfer.h>
#include <phStbRpc_Linux.h>
#include <phStbDbg.h>
#include "phStbRpc_Module.h"

DBG_UNIT_EXTERN(gphStbRpc_DbgUnit)

/***********************************************
* LOCAL MACROS                                 *
* recommendation only: <MODULE><_WORD>+        *
************************************************/

#define PH_STBRPC_NUM_TRIES 20U
#define PH_STBRPC_TRY_MSLEEP 10U

/******************************************************************
* FUNCTION PROTOTYPES                    [p]ph<Module>_<Word>+    *
*******************************************************************/

/******************************************************************
* LOCAL TYPEDEFS                                                  *
* recommendation only:                   [p]ph<Module>_<Word>+_t  *
*******************************************************************/

/*******************************************************************
* EXPORTED DATA                 g[k|p|kp|pk|kpk]ph<Module>_<Word>+ *
********************************************************************/

/*******************************************************************
* LOCAL DATA                                                       *
********************************************************************/

/*******************************************************************************
* FUNCTION IMPLEMENTATION  <Module>[_<Word>+] for static functions             *
*                          tm[<layer>]<Module>[_<Word>+] for exported functions*
********************************************************************************/

/******************************************************************************/
static void StbRpc_Xfer_Handler(LUD_COMMAND* tc, LUD_DEV* m)
{
    /* code to handle the command tc from the module m. */
    DBG_PRINT1(gphStbRpc_DbgUnit, DBG_LEVEL_5,
               "phStbRpc_Xfer_Handler: Lud command rcvd 0x%x\n", tc->command);
#if 0
    switch (tc->command)
    {
//        case :
//            break;
    }
#endif
}

/******************************************************************************/

tmErrorCode_t phStbRpc_Xfer_InitHandler(
    pphStbRpc_Xfer_t pXfer,
    phStbRpc_TargetId_t targetId,
    phStbRpc_HandlerId_t handlerId)
{
    sprintf(pXfer->device, "rpc%u-%u", targetId, handlerId);
    DBG_PRINT3(gphStbRpc_DbgUnit, DBG_LEVEL_1,
               "phStbRpc_Xfer_Create: handler=%u-%s opening \"%s\"\n",
               targetId, gphStbRpc_HandlerNames[handlerId], pXfer->device);
    DBG_ASSERT(pXfer->lastFunctionId == PH_STBRPC_XFER_UNINITIALISED);

    pXfer->link = lud_get_node_link(NULL, pXfer->device, StbRpc_Xfer_Handler);
    if (pXfer->link == NULL)
    {
        DBG_PRINT3(gphStbRpc_DbgUnit, DBG_WARNING,
                   "phStbRpc_Xfer_Create: handler=%u-%s lud_get_node_link(\"%s\") failed\n",
                   targetId, gphStbRpc_HandlerNames[handlerId], pXfer->device);
        return PH_ERR_STBRPC_CLOSED;
    }
    pXfer->readChan = lud_new_read_channel(pXfer->link);
    pXfer->writeChan = lud_new_write_channel(pXfer->link);

    pXfer->lastFunctionId = PH_STBRPC_XFER_INITIALISED;
    return TM_OK;
}

/******************************************************************************/
/* Receive an RPC reply.
   If maxRpcBufSize is 0 then don't read any data.
   Unlocks the handler mutex.
*/
Int32 phStbRpc_Xfer_Receive(
    pphStbRpc_Xfer_t pXfer,
    pUInt8 *ppBuffer,
    UInt32 maxRpcBufSize)
{
    Int32 totalSizeToRead;
    
    if (maxRpcBufSize == 0)
    {
        totalSizeToRead = 0;
    }
    else
    {
        Int nBytesRead;
        UInt numTries = 1;
        /* Retry in case Luddite fails, which sometimes happens */
        for (;;)
        {
            for (;;)
            {
                nBytesRead = lud_read_channel(pXfer->readChan, *ppBuffer , maxRpcBufSize,
                                              !(((pphStbRpc_RpcBuf_t)*ppBuffer)->rpcFlagsOrStatus & phStbRpc_Option_UserSpace), // data is in kernel space
                                              0); // not nonblocking
                if (nBytesRead != -ERESTARTSYS)
                {
                    break;
                }
                /* Wait a little bit: */
                schedule();
            }
            
            if (nBytesRead < 0)
            {
                if (numTries == 1)
                {
                    DBG_PRINT4(gphStbRpc_DbgUnit, DBG_WARNING,
                               "phStbRpc_Xfer_Receive: handler=%u-%s read from \"%s\" => %d (retrying)\n",
                               pXfer->targetId, gphStbRpc_HandlerNames[pXfer->handlerId], pXfer->device, nBytesRead);
                }
                else
                {
                    if (numTries == PH_STBRPC_NUM_TRIES)
                    {
                        DBG_PRINT5(gphStbRpc_DbgUnit, DBG_WARNING,
                                   "phStbRpc_Xfer_Receive: handler=%u-%s read from \"%s\" => %d (giving up after %d attempts)\n",
                                   pXfer->targetId, gphStbRpc_HandlerNames[pXfer->handlerId], pXfer->device, nBytesRead, numTries);
                        goto ERROR_RETURN;
                    }
                }
                (void)msleep_interruptible(PH_STBRPC_TRY_MSLEEP);
                numTries++;
            }
            else
            {
                break;
            }
        }
            
        // Get the amount we should have received in total:
        totalSizeToRead = *(pUInt32)*ppBuffer;
        if (totalSizeToRead != nBytesRead)
        {
            DBG_PRINT5(gphStbRpc_DbgUnit, DBG_WARNING,
                       "phStbRpc_Xfer_Receive: handler=%u-%s read from \"%s\" totalSizeToRead(%d) failed => %d\n",
                       pXfer->targetId, gphStbRpc_HandlerNames[pXfer->handlerId], pXfer->device, totalSizeToRead, nBytesRead);
            goto ERROR_RETURN;
        }
        if (totalSizeToRead > maxRpcBufSize)
        {
            DBG_PRINT5(gphStbRpc_DbgUnit, DBG_WARNING,
                       "phStbRpc_Xfer_Receive: handler=%u-%s read from \"%s\" of %d bytes does not fit buffer size %u\n",
                       pXfer->targetId, gphStbRpc_HandlerNames[pXfer->handlerId], pXfer->device, totalSizeToRead, maxRpcBufSize);
            goto ERROR_RETURN;
        }
    }
    goto RETURN;
    
  ERROR_RETURN:
    totalSizeToRead = -1;

  RETURN:
    up(&pXfer->mutex);
    return totalSizeToRead;
}

/******************************************************************************/
/* Send a RPC request.
   First locks the handler mutex (which may already be held by the current process).
   On error, it unlocks the mutex (since the rpc call will abort).
 */
tmErrorCode_t phStbRpc_Xfer_Send(
    pphStbRpc_Xfer_t pXfer,
    pUInt8 pBuffer,
    UInt32 bufferSize)
{
    ssize_t len;
    pphStbRpc_RpcBuf_t pRpcBuf = (pphStbRpc_RpcBuf_t)pBuffer;
    int kernelMode = !(pRpcBuf->rpcFlagsOrStatus & phStbRpc_Option_UserSpace);
    UInt numTries = 1;
    
    /* The Xfer may already be locked by StbRpc_XferLookup: */
    if (kernelMode)
    {
        if (down_interruptible(&pXfer->mutex) != 0)
        {
            return PH_ERR_STBRPC_INTERRUPTED;
        }
    }
    
    pXfer->nCalls++;
    pXfer->lastFunctionId = pRpcBuf->funIdOrResult;

    /* Retry in case Luddite fails, which sometimes happens */
    for (;;)
    {
        len = lud_write_channel_ex(pXfer->writeChan, pBuffer, bufferSize,
                                   kernelMode, // data is in kernel space
                                   0, // not nonblocking
                                   1  // flush
                                   );
            
        if (len != bufferSize)
        {
            if (numTries == 1)
            {
                DBG_PRINT5(gphStbRpc_DbgUnit, DBG_WARNING,
                           "phStbRpc_Xfer_Send: handler=%u-%s write to \"%s\" bytes=%ld => %d (retrying)\n",
                           pXfer->targetId, gphStbRpc_HandlerNames[pXfer->handlerId], pXfer->device, bufferSize, len);
            }
            else
            {
                if (numTries == PH_STBRPC_NUM_TRIES)
                {
                    DBG_PRINT6(gphStbRpc_DbgUnit, DBG_WARNING,
                               "phStbRpc_Xfer_Send: handler=%u-%s write to \"%s\" bytes=%ld failed (giving up after %d attempts)\n",
                               pXfer->targetId, gphStbRpc_HandlerNames[pXfer->handlerId], pXfer->device, bufferSize, len, numTries);
                    goto ERROR_RETURN;
                }
            }
            (void)msleep_interruptible(PH_STBRPC_TRY_MSLEEP);
            numTries++;
        }
        else
        {
            break;
        }
    }

    schedule();
    return TM_OK;
    
  ERROR_RETURN:
    up(&pXfer->mutex);
    return PH_ERR_STBRPC_XFER;
}

/******************************************************************************/
