/*  Copyright (C) 2004 Koninklijke Philips Electronics N.V.
    All Rights Reserved.
*//**
    @file   phStbRpc_Module.c
    @brief  Linux kernel module for phStbRpc component.

@b  Component:  phStbRpc.

*//*

Rev Date        Author      Comments
--------------------------------------------------------------------------------
  1 20050512    batelaan    Initial
  2 20050523    batelaan    Crosscheck fixes
  3 20050531    batelaan    Updated copyright.
  4 20050701    batelaan    Implemented properly.
  5 20050705    batelaan    Crosscheck fixes
  6 20050712    batelaan    Add phStbRpc_ProxyDeInit
  7 20050808    batelaan    phStbRpc_Xfer ifce changes.
  8 20050906    batelaan    Enable second TM.
  9 20051003    batelaan    Add phStbRpc_Xfer_Init
 10 20051102    batelaan    Add ioctl for doing rpc from usermode.
                            Add /proc entry for rpc call counting.
 11 20051104    batelaan    Crosscheck changes.
 12 20051107    batelaan    Retry on Luddite channel error.
 13 20051107    batelaan    Comparison corrections.
 14 20051107    batelaan    Retry forever while ERESTARTSYS is returned.
 15 20051109    batelaan    Split out luddite code.
 16 20051116    burningh    Ensure phStbRpc can be reloaded
 17 20051121    laird       Allow RPC to b de-inited before it gets unloaded.
 18 20051130    laird       XCheck rework.
 19 20051123    raiyat      Linux 2.6.14.2 changes
 20 20060908    laird       Linux 2.6.17.7 changes
--------------------------------------------------------------------------------
    For consistency and standardisation retain the Section Separators.
*/

/* needed for versioned kernels (i.e. most kernels in current use) */
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
//#include <linux/proc_fs.h>
#include <linux/slab.h>        /* kmalloc() */
#include <linux/errno.h>
#include <linux/fs.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <linux/delay.h>

#include <linux/device.h>
#include <linux/cdev.h>

#include <phStbRpc.h>
#include <phStbRpc_Internal.h>
#include <phStbRpc_Xfer.h>
#include <phStbRpc_Linux.h>
#include <phStbDbg.h>
#include "phStbRpc_Module.h"
#include <phStbProc_BaseProcSystem.h>

DBG_UNIT_EXTERN(gphStbRpc_DbgUnit)

/***********************************************
* LOCAL MACROS                                 *
* recommendation only: <MODULE><_WORD>+        *
************************************************/

#define MY_LICENSE "GPL"
MODULE_LICENSE(MY_LICENSE);

#define THIS_MODULE_DESCRIPTION "TriMedia RPC driver"
MODULE_DESCRIPTION(THIS_MODULE_DESCRIPTION);
MODULE_AUTHOR("Wouter Batelaan");

/** @see StbRpc_ProcRead */
#define PH_STBRPC_CREATE_PROC_ENTRIES 1

/* The number of devices we expose */
#define NR_DEVS 1

#define PH_STBRPC_NUM_TRIES 20U
#define PH_STBRPC_TRY_MSLEEP 10U

/******************************************************************
* FUNCTION PROTOTYPES                    [p]ph<Module>_<Word>+    *
*******************************************************************/

static int __init StbRpc_ModuleInit(void);
static void __exit StbRpc_ModuleExit(void);
static int StbRpc_open(struct inode *inode, struct file *file);
static ssize_t StbRpc_release(struct inode *inode, struct file *file);
static int StbRpc_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int StbRpc_ProcRead(
    char *buffer,
    char **buffer_location,
    off_t offset, int buffer_length, int *eof, void *data);

/******************************************************************
* LOCAL TYPEDEFS                                                  *
* recommendation only:                   [p]ph<Module>_<Word>+_t  *
*******************************************************************/

/*******************************************************************
* EXPORTED DATA                 g[k|p|kp|pk|kpk]ph<Module>_<Word>+ *
********************************************************************/

module_init(StbRpc_ModuleInit);
module_exit(StbRpc_ModuleExit);

/*******************************************************************
* LOCAL DATA                                                       *
********************************************************************/

/** The Xfers for each handler. */
static phStbRpc_Xfer_t gStbRpc_Xfers[phStbRpc_TargetId_Last - phStbRpc_TargetId_First + 1][phStbRpc_HandlerId_Count];

static dev_t gDev; /* The start of the reserved device number region */
static struct cdev gCdev;
static struct class *gSysClass; /* The device class under /sys/class */
static unsigned int gMajor; /* Major number of our device, obtained from gDev and held for convenience */

/* The file operations for the device */
static struct file_operations phStbRpc_fops =
{
    .owner      = THIS_MODULE,
    .open       = StbRpc_open,
    .release    = StbRpc_release,
    .ioctl      = StbRpc_ioctl
};

/*******************************************************************************
* FUNCTION IMPLEMENTATION  <Module>[_<Word>+] for static functions             *
*                          tm[<layer>]<Module>[_<Word>+] for exported functions*
********************************************************************************/

/******************************************************************************/
/** Module Creation function.
    - creates the philips/rpc proc entry directory if PH_STBRPC_CREATE_PROC_ENTRIES!=0.

    @retval int Module Initialised successfully.
*/
static int __init StbRpc_ModuleInit(void)
{
    int result;

    /** To track module initialisation progress, so we can undo correctly on error. */
    enum {
        cdevAdded = 1<<0,
        classSimpleDeviceAdded = 1<<1
    } initialised = 0;

    /* Allocate a region of minor numbers (1) under a major number */
    result = alloc_chrdev_region(&gDev, 0, NR_DEVS, PH_STBRPC_DEVICE_NAME);
    if (result < 0)
    {
        printk(KERN_WARNING "phStbRpc: alloc_chrdev_region => %d\n", result);
        goto ERROR_RETURN;
    }

    gMajor = MAJOR(gDev);

    /* Set up /sys entry */
    gSysClass = class_create(THIS_MODULE, PH_STBRPC_DEVICE_NAME);
    if (gSysClass == NULL)
    {
        printk(KERN_WARNING "phStbRpc: class_create => %d\n", result);
        result = -ENOMEM;
        goto ERROR_RETURN;
    }

    cdev_init(&gCdev, &phStbRpc_fops);
    gCdev.owner = THIS_MODULE;

    result = cdev_add(&gCdev, gDev, 1);
    if (result)
    {
        printk(KERN_INFO "phStbRpc: cdev_add => %d\n", result);
        goto ERROR_RETURN;
    }
    initialised |= cdevAdded;

    class_device_create(gSysClass, NULL, gDev, NULL, PH_STBRPC_DEVICE_NAME);
    initialised |= classSimpleDeviceAdded;

#if PH_STBRPC_CREATE_PROC_ENTRIES
    {
        struct proc_dir_entry *philips_base_dir;
        char name[10];
        phStbRpc_TargetId_t t;
        result = phStbProc_GetPhilipsProcEntry(&philips_base_dir);
        if (result != 0)
        {
            goto ERROR_RETURN;
        }

        for (t = phStbRpc_TargetId_First; t <= phStbRpc_TargetId_Last; t++)
        {
            sprintf(name, "rpc%u", t);
            (void)create_proc_read_entry(name, 0444, philips_base_dir, &StbRpc_ProcRead, (void*)t);
        }
    }
#endif
    phStbRpc_Xfer_Init();

    printk ("%s (%s-%s)\n", THIS_MODULE_DESCRIPTION, __DATE__, __TIME__);
    return 0;

  ERROR_RETURN:
    printk(KERN_WARNING "%s (%s-%s) => %d\n", THIS_MODULE_DESCRIPTION, __DATE__, __TIME__, result);
    if (initialised & classSimpleDeviceAdded)
    {
        class_device_destroy(gSysClass, gDev);
    }

    if (initialised & cdevAdded)
    {
        cdev_del(&gCdev);
    }

    if (gSysClass != NULL)
    {
        class_destroy(gSysClass);
    }

    return result;
}


/******************************************************************************/
/** Remove the proc entries if PH_STBRPC_CREATE_PROC_ENTRIES!= 0
    and de-init the module
*/
static void __exit StbRpc_ModuleExit(void)
{
    printk("Shutting down %s\n", THIS_MODULE_DESCRIPTION);

#if PH_STBRPC_CREATE_PROC_ENTRIES
    {
        struct proc_dir_entry *philips_base_dir;
        char name[10];
        phStbRpc_TargetId_t t;
        phStbProc_GetPhilipsProcEntry(&philips_base_dir);

        for (t = phStbRpc_TargetId_First; t <= phStbRpc_TargetId_Last; t++)
        {
            sprintf(name, "rpc%u", t);
            remove_proc_entry(name, philips_base_dir);
        }
    }
#endif

    class_device_destroy(gSysClass, gDev);
    class_destroy(gSysClass);
    cdev_del(&gCdev);
}

/******************************************************************************/
/**
 * Read the /proc file for a target.
 * The format of the proc file is:
 * <pre>
 * handlerId TAB numberOfCalls TAB lastOrCurrentFunctionId TAB handlerName
 * </pre>
 */
static
int
StbRpc_ProcRead(char *buffer,
                char **buffer_location,
                off_t offset, int buffer_length, int *eof, void *data)
{
    int ret;
    phStbRpc_TargetId_t targetId = (phStbRpc_TargetId_t)data;
    if (offset > 0) {
        /* we have finished to read, return 0 */
        ret  = 0;
    } else {
        phStbRpc_HandlerId_t h;
        for (h = 0; h  < phStbRpc_HandlerId_Count; h++)
        {
            pphStbRpc_Xfer_t pXfer = &gStbRpc_Xfers[targetId - phStbRpc_TargetId_First][h];
            offset += snprintf(buffer + offset, buffer_length - offset,
                               "%u\t%u\t%d\t%s\n", h, pXfer->nCalls, pXfer->lastFunctionId, gphStbRpc_HandlerNames[h]);
        }
    }

    return offset;
}

/******************************************************************************/
/** Open the device.
 */
static int StbRpc_open(struct inode *inode, struct file *file)
{
    return 0;
}

/******************************************************************************/
/** Close the device.
 */
static ssize_t StbRpc_release(struct inode *inode, struct file *file)
{
    return 0;
}

/******************************************************************************/
tmErrorCode_t phStbRpc_Xfer_Lookup(
    UInt32 targetId,
    UInt32 handlerId,
    pphStbRpc_Xfer_t *ppXfer)
{
    pphStbRpc_Xfer_t pXfer;
    if (!PH_STBRPC_OK_TARGETANDHANDLERID(targetId, handlerId))
    {

        return PH_ERR_STBRPC_HANDLERID;
    }

    pXfer = &gStbRpc_Xfers[targetId - phStbRpc_TargetId_First][handlerId];

    if (down_interruptible(&pXfer->mutex) != 0)
    {
        return PH_ERR_STBRPC_INTERRUPTED;
    }

    /* Initialise? */
    if (pXfer->lastFunctionId == PH_STBRPC_XFER_UNINITIALISED)
    {
        tmErrorCode_t err = phStbRpc_Xfer_InitHandler(pXfer, targetId, handlerId);
        if (err != TM_OK)
        {
            up(&pXfer->mutex);
            return err;
        }
    }
    *ppXfer = pXfer;
    return TM_OK;
}

/******************************************************************************/
/** Handle rpc call from user mode.
    @return A negative error code on error, or non-negative number of bytes received if
    RPC calls succeeds. */
static int StbRpc_UserModeRpcCall(pphStbRpc_RpcBuf_t pRpcBuf)
{
    pphStbRpc_Xfer_t pXfer;
    tmErrorCode_t err;
    UInt32 targetId;
    UInt32 handlerId;
    UInt32 len;

#if 0
    /* Can't get this to compile: */
    if (!(access_ok(VERIFY_WRITE, PH_STBRPC_RPCBUF_OVERHEAD, pRpcBuf) &&
          access_ok(VERIFY_WRITE, pRpcBuf->size, pRpcBuf)))
    {
        return -EFAULT;
    }
#endif

    targetId = PH_STBRPC_GET_TARGETID(pRpcBuf->handlerId);
    handlerId = PH_STBRPC_GET_HANDLERID(pRpcBuf->handlerId);

    err = phStbRpc_Xfer_Lookup(targetId, handlerId, &pXfer);
    if (err != TM_OK)
    {
        return -err;
    }

    pRpcBuf->rpcFlagsOrStatus |= phStbRpc_Option_UserSpace;
    err = phStbRpc_Xfer_Send(pXfer, (pUInt8)pRpcBuf, pRpcBuf->size);
    if (err != TM_OK)
    {
        return -err;
    }

    len = phStbRpc_Xfer_Receive(pXfer, (pUInt8*)&pRpcBuf,
                                (pRpcBuf->rpcFlagsOrStatus & phStbRpc_Option_Async) ? 0 : pRpcBuf->size);
    if (len >= 0)
    {
        return len;
    }
    return -PH_ERR_STBRPC_XFER;

}

/******************************************************************************/
/* Handle all ioctls supported by phStbRpc */
static int StbRpc_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    int result = -ENOTTY;

    switch (cmd)
    {
        case PH_STBRPC_IOCTL_RPCCALL:
        {
            pphStbRpc_RpcBuf_t pRpcBuf = (pphStbRpc_RpcBuf_t)arg;
            result = StbRpc_UserModeRpcCall(pRpcBuf);
        }
        break;

        default:
            break;
    }

    return result;
}

/******************************************************************************/
/* Creates or looks up the Xfer. The Xfer mutex lock won't be held after this function.
   This is because phStbRpc_Xfer_Create is used from kernelmode only,
   and may not be followed by an rpc call for a while.
   When the kernelmode rpc user calls phStbRpc_Xfer_Send, we will
   lock the mutex anyway.
 */
tmErrorCode_t phStbRpc_Xfer_Create(
    phStbRpc_TargetId_t targetId,
    phStbRpc_HandlerId_t handlerId,
    UInt32 maxBufferSize,
    pphStbRpc_Xfer_t *ppXfer)
{
    tmErrorCode_t err;
    pphStbRpc_Xfer_t pXfer;

    err = phStbRpc_Xfer_Lookup(targetId, handlerId, &pXfer);
    if (err != TM_OK)
    {
        return err;
    }

    *ppXfer = pXfer;
    up(&pXfer->mutex);

    return TM_OK;
}

/******************************************************************************/

Void phStbRpc_Xfer_Destroy(
    pphStbRpc_Xfer_t pXfer)
{
    /* Since the Xfers are stored statically and can be used also from user-mode,
       we never really free them. */
}

/******************************************************************************/

void * phStbRpc_Xfer_Malloc(
    pphStbRpc_Xfer_t pXfer,
    UInt32 size)
{
    void *p = kmalloc(size, GFP_KERNEL);
#if TMDBG_PRINTS_ARE_ENABLED
    if (p == NULL)
    {
        DBG_PRINT3(gphStbRpc_DbgUnit, DBG_WARNING,
                   "phStbRpc_Xfer_Malloc: handler=%u-%s size=%u failed\n",
                   pXfer->targetId, gphStbRpc_HandlerNames[pXfer->handlerId], size);
    }
#endif
    return p;
}

/******************************************************************************/

void phStbRpc_Xfer_Free(
    pphStbRpc_Xfer_t pXfer,
    void *address)
{
    kfree(address);
}

/******************************************************************************/

void phStbRpc_Xfer_Init(Void)
{
    phStbRpc_TargetId_t targetId;
    for (targetId = phStbRpc_TargetId_First; targetId <= phStbRpc_TargetId_Last; targetId++)
    {
        phStbRpc_HandlerId_t handlerId;
        for (handlerId = 0; handlerId  < phStbRpc_HandlerId_Count; handlerId++)
        {
            pphStbRpc_Xfer_t pXfer = &gStbRpc_Xfers[targetId - phStbRpc_TargetId_First][handlerId];
            init_MUTEX(&pXfer->mutex);
            pXfer->targetId = targetId;
            pXfer->handlerId = handlerId;
            pXfer->lastFunctionId = PH_STBRPC_XFER_UNINITIALISED;
        }
    }
}

/******************************************************************************/

void phStbRpc_Xfer_DeInit(Void)
{
    phStbRpc_TargetId_t targetId;
    for (targetId = phStbRpc_TargetId_First; targetId <= phStbRpc_TargetId_Last; targetId++)
    {
        phStbRpc_HandlerId_t handlerId;
        for (handlerId = 0; handlerId  < phStbRpc_HandlerId_Count; handlerId++)
        {
            pphStbRpc_Xfer_t pXfer = &gStbRpc_Xfers[targetId - phStbRpc_TargetId_First][handlerId];
            pXfer->lastFunctionId = PH_STBRPC_XFER_UNINITIALISED;
        }
    }
}

/******************************************************************************/
