/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright (C) 2006 NXP B.V.
 * All Rights Reserved.
*//**
    @file phStbRpc_Proxy.c
    Proxy for RPC.

    @b  Component:  StbRpc.

    File Name = phStbRpc_Proxy.c
*//*

Rev Date       Author      Comments
--------------------------------------------------------------------------------
  1 20050419   batelaan    Initial
  2 20050511   batelaan    Make it work properly.
  3 20050512   batelaan    Crosscheck fixes.
  4 20050621   batelaan    Char->UInt8; add error msg.
  5 20050630   batelaan    Correct error handling.
  6 20050704   batelaan    Enable compile as a module.
  7 20050705   batelaan    Crosscheck fixes
  8 20050712   batelaan    Add phStbRpc_ProxyDeInit
  9 20050719   batelaan    Call phStbRpc_ProxyInit if needed within phStbRpc_RpcCall
 10 20050808   batelaan    phStbRpc_Xfer ifce changes.
 11 20050810   batelaan    Added payloadMaxLen param to phStbRpc_ProxyInit
 12 20050811   batelaan    Compile errors on WinCE.
 13 20050927   batelaan    unistd.h not needed (and not available on WinCE)
 16 20051025   batelaan    PR 247: automatic proxy init with wrong payload size.
 17 20051102   batelaan    Add gphStbRpc_HandlerNames
 18 20051121   laird       Allow RPC to be de-inited before it gets unloaded.
 19 20051130   laird       XCheck rework.
 20 20051130   laird       XCheck rework.
 21 20060127   batelaan    Add check for PH_STBRPC_HANDLERID_NAMES vs phStbRpc_HandlerId_Count
 22 20070730    laird       Updated Header
*/


/* Do not remove the following comments even if not applicable.
   This makes the fact explicit that for eg. there are no local typedefs,
   and allows consistant layout during maintenance phases     */

/*******************
* INCLUDE FILES    *
********************/

#if defined(__linux__) && defined(MODULE)
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/limits.h>
#else
#include <stdio.h>
#include <memory.h>
#include <limits.h>
#endif

#include <phStbCommon.h>
#include <phStbDbg.h>
#include <phStbRpc.h>
#include <phStbRpc_Xfer.h>

/*******************
* LOCAL MACROS     *
********************/

#define PH_STBRPC_PROXY_MAX_ENTRIES (INT_MAX/2)

/*******************
* EXPORTED DATA    *
********************/

DBG_UNIT(gphStbRpc_DbgUnit)

const char *gphStbRpc_HandlerNames[] = {PH_STBRPC_HANDLERID_NAMES};

/*******************
* LOCAL TYPEDEFS   *
********************/

/** Information for a Rpc handler on a host CPU.
    @see phStbRpc_ProxyInit().
*/
typedef struct phStbRpc_Proxy {
    /** Id of this Rpc handler. */
    phStbRpc_HandlerId_t handlerId;
    /** For the data transfer. */
    pphStbRpc_Xfer_t pXfer;
    /** Number of Rpc functions that this handler has. */
    UInt32 numFun;
} phStbRpc_Proxy_t;

/** Pointer to a ::phStbRpc_Proxy_t. */
typedef phStbRpc_Proxy_t *pphStbRpc_Proxy_t;

/**********************
* FUNCTION PROTOTYPES *
***********************/

/*******************
* STATIC DATA      *
********************/

static phStbRpc_Proxy_t gStbRpc_Proxies[phStbRpc_TargetId_Last - phStbRpc_TargetId_First + 1]
                                       [phStbRpc_HandlerId_Count];

static Bool gStbRpc_Initialised = 0;

/***************************
* FUNCTION IMPLEMENTATIONS *
****************************/

/******************************************************************************/
static void StbRpc_CheckInitialised(Void)
{
    if (!gStbRpc_Initialised)
    {
        DBG_REGISTER_MODULE(gphStbRpc_DbgUnit, "StbRpc");
        /* If the following assert fails then there is a mismatch between the definitions
           of phStbRpc_HandlerId_t and PH_STBRPC_HANDLERID_NAMES. */
        DBG_ASSERT((sizeof(gphStbRpc_HandlerNames)/sizeof(gphStbRpc_HandlerNames[0]))
                   == phStbRpc_HandlerId_Count);
        gStbRpc_Initialised = 1;
        phStbRpc_Xfer_Init();
    }
}

/******************************************************************************/
void phStbRpc_Init(Void)
{
    StbRpc_CheckInitialised();
}

/******************************************************************************/
void phStbRpc_DeInit(Void)
{
    int targetId = 0;
    int handlerId = 0;

    StbRpc_CheckInitialised();

    for( targetId = phStbRpc_TargetId_First; targetId <= phStbRpc_TargetId_Last; targetId++)
    {
        for(handlerId =0; handlerId < phStbRpc_HandlerId_Count; handlerId++)
        {
            pphStbRpc_Proxy_t pProxy;

            pProxy = &gStbRpc_Proxies[targetId][handlerId];

            if (pProxy->pXfer != NULL)
            {
                phStbRpc_Xfer_Destroy(pProxy->pXfer);
                pProxy->pXfer = NULL;
            }
        }
    }

    phStbRpc_Xfer_DeInit();
    gStbRpc_Initialised = 0;
}

/******************************************************************************/
tmErrorCode_t phStbRpc_ProxyInit(
    phStbRpc_HandlerId_t handlerId,
    UInt32 numFun,
    UInt32 maxPayloadSize
)
{
    pphStbRpc_Proxy_t pProxy;
    phStbRpc_TargetId_t targetId = PH_STBRPC_GET_TARGETID(handlerId);
    handlerId = PH_STBRPC_GET_HANDLERID(handlerId);

    StbRpc_CheckInitialised();
    if (!PH_STBRPC_OK_TARGETANDHANDLERID(targetId, handlerId))
    {
        DBG_PRINT2(gphStbRpc_DbgUnit, DBG_WARNING,
                   "phStbRpc_ProxyInit: illegal id %u-%u", targetId, handlerId);
        return PH_ERR_STBRPC_HANDLERID;
    }

    DBG_PRINT4(gphStbRpc_DbgUnit, DBG_LEVEL_1,
               "phStbRpc_ProxyInit handler=%u-%s numFun=%u maxPayloadSize=%u",
               targetId, gphStbRpc_HandlerNames[handlerId], numFun, maxPayloadSize);

    pProxy = &gStbRpc_Proxies[targetId - phStbRpc_TargetId_First][handlerId];

    if (pProxy->pXfer == NULL)
    {
        tmErrorCode_t err;
        err = phStbRpc_Xfer_Create(targetId, handlerId, PH_STBRPC_RPCBUF_SIZE(maxPayloadSize), &pProxy->pXfer);
        if (err != 0)
        {
            return err;
        }
        pProxy->handlerId = handlerId;
        pProxy->numFun = numFun;
    }
    return 0;
}

/******************************************************************************/
tmErrorCode_t phStbRpc_ProxyDeInit(
    phStbRpc_HandlerId_t handlerId)
{
    pphStbRpc_Proxy_t pProxy;
    phStbRpc_TargetId_t targetId = PH_STBRPC_GET_TARGETID(handlerId);
    handlerId = PH_STBRPC_GET_HANDLERID(handlerId);

    StbRpc_CheckInitialised();

    if (!PH_STBRPC_OK_TARGETANDHANDLERID(targetId, handlerId))
    {
        DBG_PRINT2(gphStbRpc_DbgUnit, DBG_WARNING,
                   "phStbRpc_ProxyDeInit: illegal id %u-%u", targetId, handlerId);
        return PH_ERR_STBRPC_HANDLERID;
    }

    DBG_PRINT2(gphStbRpc_DbgUnit, DBG_LEVEL_1,
               "phStbRpc_ProxyDeInit handler=%u-%s",
               targetId, gphStbRpc_HandlerNames[handlerId]);

    pProxy = &gStbRpc_Proxies[targetId - phStbRpc_TargetId_First][handlerId];

    if (pProxy->pXfer == NULL)
    {
        DBG_PRINT2(gphStbRpc_DbgUnit, DBG_WARNING,
                   "phStbRpc_ProxyDeInit handler=%u-%s => PH_ERR_STBRPC_NOT_INIT",
                   targetId, gphStbRpc_HandlerNames[handlerId]);
        return PH_ERR_STBRPC_NOT_INIT;
    }
    phStbRpc_Xfer_Destroy(pProxy->pXfer);
    pProxy->pXfer = NULL;
    return 0;
}

/******************************************************************************/
tmErrorCode_t phStbRpc_RpcCall(
    phStbRpc_HandlerId_t handlerId,
    UInt32 funId,
    UInt32 flags,
    pUInt8 pBuffer,
    pUInt32 pBufferSize,
    pUInt32 pResult
)
{
    pphStbRpc_Proxy_t pProxy;
    tmErrorCode_t err;
    pphStbRpc_RpcBuf_t pRpcBuf;
    UInt32 rpcBufSize;
    Int len;
    UInt32 bufferSize = *pBufferSize;
    phStbRpc_TargetId_t targetId = PH_STBRPC_GET_TARGETID(handlerId);
    handlerId = PH_STBRPC_GET_HANDLERID(handlerId);

    StbRpc_CheckInitialised();

    DBG_ASSERT(pBuffer != NULL);
    DBG_ASSERT(pBufferSize != NULL);

    if (!PH_STBRPC_OK_TARGETANDHANDLERID(targetId, handlerId))
    {
        DBG_PRINT2(gphStbRpc_DbgUnit, DBG_WARNING,
                   "phStbRpc_RpcCall handler=%u-%u => PH_ERR_STBRPC_HANDLERID",
                   targetId, handlerId);
        return PH_ERR_STBRPC_HANDLERID;
    }

    DBG_PRINT5(gphStbRpc_DbgUnit, DBG_LEVEL_2,
               "phStbRpc_RpcCall handler=%u-%s funId=%u flags=0x%x bufSize=%u",
               targetId, gphStbRpc_HandlerNames[handlerId], funId, flags, *pBufferSize);

    pProxy = &gStbRpc_Proxies[targetId - phStbRpc_TargetId_First][handlerId];

    if (pProxy->pXfer == NULL)
    {
        /* Since the proxy has not been initialised, do it here automatically.
           Since we haven't got here the number of functions available,
           set it to INT_MAX. */
        err = phStbRpc_ProxyInit(PH_STBRPC_HANDLERID(targetId, handlerId),
                                 INT_MAX, *pBufferSize);
        if (err != 0)
        {
            return err;
        }
    }

    if (funId >= pProxy->numFun)
    {
        DBG_PRINT3(gphStbRpc_DbgUnit, DBG_WARNING,
                   "phStbRpc_RpcCall handler=%u-%s funId=%u => PH_ERR_STBRPC_FUNID",
                   targetId, gphStbRpc_HandlerNames[handlerId], funId);
        return PH_ERR_STBRPC_FUNID;
    }

    rpcBufSize = PH_STBRPC_RPCBUF_SIZE(bufferSize);
    pRpcBuf = phStbRpc_Xfer_Malloc(pProxy->pXfer, rpcBufSize);
    if (pRpcBuf == NULL)
    {
        return PH_ERR_STBRPC_NO_MEM;
    }

    pRpcBuf->size = rpcBufSize;
    pRpcBuf->handlerId = pProxy->handlerId;
    pRpcBuf->funIdOrResult = funId;
    pRpcBuf->rpcFlagsOrStatus = flags;
    /* Store the payload data in the rpc buffer: */
    if (bufferSize != 0)
    {
        memcpy(&pRpcBuf->buffer, pBuffer, bufferSize);
    }
    err = phStbRpc_Xfer_Send(pProxy->pXfer, (pUInt8)pRpcBuf, rpcBufSize);
    if (err == 0)
    {
        pUInt8 pBufferRcv = (pUInt8)pRpcBuf;
        // Read back a result
        len = phStbRpc_Xfer_Receive(pProxy->pXfer, &pBufferRcv,
                                   (flags & phStbRpc_Option_Async) ? 0 : rpcBufSize);
        if (!(flags & phStbRpc_Option_Async))
        {
            /* RPC success? */
            err = pRpcBuf->rpcFlagsOrStatus;
            if (err == 0)
            {
                if (len != 0)
                {
                    /* Update pBuffer: */
                    len = PH_STBRPC_RPCBUF_PAYLOAD_SIZE(pRpcBuf);
                    memcpy(pBuffer, &pRpcBuf->buffer, len);
                }
                *pBufferSize = len;
                if (!(flags & phStbRpc_Option_Async))
                {
                    if (pResult != NULL)
                    {
                        *pResult = pRpcBuf->funIdOrResult;
                    }
                }
            }
        }
    }

    phStbRpc_Xfer_Free(pProxy->pXfer, pRpcBuf);
    DBG_PRINT6(gphStbRpc_DbgUnit, DBG_LEVEL_2,
               "phStbRpc_RpcCall handler=%u-%s funId=%u flags=0x%x => bufSize=%u 0x%x",
               targetId, gphStbRpc_HandlerNames[handlerId], funId, flags, *pBufferSize, err);
    return err;
}

/******************************************************************************/
#if defined(__linux__) && defined(MODULE)
  EXPORT_SYMBOL(phStbRpc_ProxyInit);
  EXPORT_SYMBOL(phStbRpc_ProxyDeInit);
  EXPORT_SYMBOL(phStbRpc_RpcCall);
#endif

/******************************************************************************/
